{-# LANGUAGE DataKinds #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE ImpredicativeTypes #-}

module Lisa.Squeak where

import Control.Applicative ((<|>))
import Control.Monad.IO.Class (MonadIO, liftIO)
import Data.Aeson (FromJSON(parseJSON), ToJSON(toJSON), Value, (.:), (.:?), (.=), object, withObject)
import Data.Text (Text)
import Data.Time.Calendar (Day)
import Network.HTTP.Req (Option, Scheme(..), POST(..), ReqBodyJson(..), Req, Scheme, Url, defaultHttpConfig, http, https, port, jsonResponse, responseBody, req, runReq)

-- | Dependency Inversion for monads that can run requests.
-- By requiring 'HasEndpointInfo' and 'MonadIO' for functions that perform requests we don't need to import anything from Scotty.
class HasEndpointInfo m where
    getEndpointInfo :: m (Either (Url 'Http, Option scheme0) (Url 'Https, Option scheme1))

-- | Debugging instance. Makes it possible to just 'getAllLectures' etc. from GHCi.
instance HasEndpointInfo IO where
    -- getEndpointInfo = pure $ Right (https "api.squeak-test.fsmi.uni-karlsruhe.de", mempty)
    getEndpointInfo = pure $ Left (http "localhost", port 8000)

-- | Construct a GraphQL query that runs in 'm' and returns a 'result'.
mkQueryWithVars
    :: (HasEndpointInfo m, MonadIO m, ToJSON v, FromJSON result)
    => v -- ^ This field is sent as JSON in the @variables@ key.
    -> Text
    -> m (Either [GQLError] result)
mkQueryWithVars vars query = do
    endpointInfo <- getEndpointInfo
    let request = (case endpointInfo of
                    Left (httpUrl, httpOptions) -> req POST httpUrl payload jsonResponse httpOptions
                    Right (httpsUrl, httpsOptions) -> req POST httpsUrl payload jsonResponse httpsOptions)
    response <- liftIO $ runReq defaultHttpConfig request
    pure $ replyToEither $ responseBody response
    where
        payload = ReqBodyJson $ GQLQuery query vars

-- | Construct a GraphQL query that runs in 'm' and returns a 'result' or a 'SqueakError'.
mkQuery
    :: (HasEndpointInfo m, MonadIO m, FromJSON result)
    => Text
    -> m (Either [GQLError] result)
mkQuery = mkQueryWithVars $ object []

joinErrors :: Either [GQLError] (SqueakErrorOr a) -> Either SqueakError a
joinErrors (Left errors) = Left $ GQLErrors errors
joinErrors (Right (SqueakErrorOr inner)) = inner

-- ^ A GraphQL query with a bunch of variables.
data GQLQuery v = GQLQuery Text v
    deriving (Show)

instance ToJSON v => ToJSON (GQLQuery v) where
    toJSON (GQLQuery query variables) = object
        [ "query" .= query
        , "variables" .= variables
        ]

data GQLReply a = GQLReply
    { gqlReplyData :: Maybe a
    , gqlReplyErrors :: Maybe [GQLError]
    }
    deriving (Show)

instance FromJSON a => FromJSON (GQLReply a) where
    parseJSON = withObject "GQLReply" $ \v -> GQLReply
        <$> v .: "data"
        <*> v .:? "errors"

replyToEither :: GQLReply a -> Either [GQLError] a
replyToEither (GQLReply (Just data') _)  = Right data'
replyToEither (GQLReply _ (Just errors)) = Left errors
replyToEither _                          = error "Invalid GQLReply: Neither 'data' nor 'errors'"

data GQLError = GQLError
    { gqlErrorMessage :: Text
    }
    deriving (Show)

instance FromJSON GQLError where
    parseJSON = withObject "GQLError" $ \v -> GQLError
        <$> v .: "message"

data Documents = Documents { unDocuments :: [Document] }
    deriving (Show)

instance FromJSON Documents where
    parseJSON = withObject "Documents" $ \v -> Documents
        <$> ((v .: "documents") >>= (.: "results"))

data Document = Document
    { documentId            :: Text
    , documentDate          :: Day
    , documentSemester      :: Text
    , documentPublicComment :: Maybe Text
    , documentFaculty       :: Faculty
    , documentLectures      :: [Lecture]
    , documentDownloadable  :: Bool
    }
    deriving (Show)

instance FromJSON Document where
    parseJSON = withObject "Document" $ \v -> Document
        <$> v .: "id"
        <*> v .: "date"
        <*> v .: "semester"
        <*> v .: "publicComment"
        <*> v .: "faculty"
        <*> v .: "lectures"
        <*> v .: "downloadable"

data Faculty = Faculty
    { facultyId          :: Text
    , facultyDisplayName :: Text
    }
    deriving (Show)

instance FromJSON Faculty where
    parseJSON = withObject "Faculty" $ \v -> Faculty
        <$> v .: "id"
        <*> v .: "displayName"

data Lectures = Lectures { unLectures :: [Lecture] }
    deriving (Show)

instance FromJSON Lectures where
    parseJSON = withObject "Lectures" $ \v -> Lectures
        <$> v .: "lectures"

data Lecture = Lecture
    { lectureId          :: Text
    , lectureDisplayName :: Text
    , lectureAliases     :: [Text]
    }
    deriving (Show)

instance FromJSON Lecture where
    parseJSON = withObject "Lecture" $ \v -> Lecture
        <$> v .: "id"
        <*> v .: "displayName"
        <*> v .: "aliases"

getAllLectures :: (HasEndpointInfo m, MonadIO m) => m (Either SqueakError [Lecture])
getAllLectures = fmap unLectures <$> joinErrors <$> mkQuery
    "{ lectures \
    \  { id displayName aliases } \
    \}"

getAllDocuments :: (HasEndpointInfo m, MonadIO m) => m (Either SqueakError [Document])
getAllDocuments = fmap unDocuments <$> joinErrors <$> mkQuery
    "{ documents(filters: []) \
    \  { results \
    \    { id date semester publicComment downloadable faculty { id displayName } lectures { id displayName aliases } } \
    \  } \
    \}"

-- Mutations

data LoginResult = LoginResult { unLoginResult :: SqueakErrorOr Credentials }
    deriving (Show)

instance FromJSON LoginResult where
    parseJSON = withObject "LoginResult" $ \v -> LoginResult
        <$> v .: "login"

data Credentials = Credentials
    { credentialsToken :: Text
    , credentialsUser  :: User
    } deriving (Show)

instance FromJSON Credentials where
    parseJSON = withObject "Credentials" $ \v -> Credentials
        <$> v .: "token"
        <*> v .: "user"

data User = User
    { userUsername    :: Text
    , userDisplayName :: Text
    } deriving (Show)

instance FromJSON User where
    parseJSON = withObject "User" $ \v -> User
        <$> v .: "username"
        <*> v .: "displayName"

{-
login :: Text -> Text -> Req (Either [GQLError] Credentials)
login username password = fmap unLoginResult <$> replyToEither <$> responseBody <$> req POST serverUrl (ReqBodyJson $ GQLQueryWithVars q vars) jsonResponse mempty
    where q = "mutation LoginUser($username: String!, $password: String!) { login(username: $username, password: $password) { ... on Credentials { token user { username displayName } } } }"
          vars = object
            [ "username" .= username
            , "password" .= password
            ]
-}

login :: (HasEndpointInfo m, MonadIO m) => Text -> Text -> m (Either SqueakError Credentials)
login username password = joinErrors <$> fmap unLoginResult <$> mkQueryWithVars
    (object ["username" .= username, "password" .= password])
    "mutation LoginUser($username: String!, $password: String!) \
    \ { login(username: $username, password: $password) \
    \    { __typename \
    \      ... on Error { errorCode msg } \
    \      ... on Credentials { token user { username displayName } } } \
    \ }"

data SqueakError
    = SqueakError
        Text -- ^ Error code
        Text -- ^ Human-readable message
    | GQLErrors [GQLError]
    deriving (Show)

instance FromJSON SqueakError where
    parseJSON = withObject "SqueakError" $ \v -> SqueakError
        <$> v .: "errorCode"
        <*> v .: "msg"

newtype SqueakErrorOr a = SqueakErrorOr (Either SqueakError a)
    deriving (Show)

instance FromJSON a => FromJSON (SqueakErrorOr a) where
    -- | Parse 'SqueakError' if @errorCode@ and @msg@ are present; otherwise, parse an 'a'.
    parseJSON v = SqueakErrorOr
        <$> ((Left <$> parseJSON v)
        <|> (Right <$> parseJSON v))

data CreateOrderResult = CreateOrderResult { unCreateOrderResult :: SqueakErrorOr Order }
    deriving (Show)

instance (FromJSON CreateOrderResult) where
    parseJSON = withObject "CreateOrderResult" $ \v -> CreateOrderResult
        <$> v .: "createOrder"

data Order = Order
    { orderId       :: Text
    , orderTag      :: Text
    , orderNumPages :: Int
    }
    deriving (Show)

instance FromJSON Order where
    parseJSON = withObject "Order" $ \v -> Order
        <$> v .: "id"
        <*> v .: "tag"
        <*> v .: "numPages"

createOrder :: (HasEndpointInfo m, MonadIO m) => Text -> [Text] -> m (Either SqueakError Order)
createOrder tag documents = joinErrors <$> fmap unCreateOrderResult <$> mkQueryWithVars
    (object ["tag" .= tag, "documents" .= documents])
    "mutation CreateOrder($tag: String!, $documents: [DocumentId!]!) \
    \ { createOrder(tag: $tag, documents: $documents) \
    \ { ... on Error { errorCode msg } \
    \   ... on Order { id tag numPages } \
    \ } }"
